# lcreg: Efficent rigid and affine 3D image registration
# 
# Copyright (C) 2019  Peter Rösch, Peter.Roesch@hs-augsburg.de
# 
# Organisation:
# Faculty of Computer Science, Augsburg University of Applied Sciences,
# An der Hochschule 1, 86161 Augsburg, Germany
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

#cython: language_level=3, boundscheck=False, profile=False, cdivision=True, wraparound=False
cimport numpy as np
import numpy as np

cimport cython
cimport openmp

from cython.parallel import prange, parallel
from libc.stdlib cimport malloc, free, abort
cimport lcreg.lcreg_lib

cdef double mirrored_interpolate_1d(double *a, index_t N, double x) nogil:
    """ Calculate interpolated grey values in a 1D float array.

        Args:
            a: 1D numpy float64 array
            N: length of a
            x: position at which the gray value should be interpolated
        Returns:
            interpolated value
    """
    # check if position is within the array including border pixels
    if x <= -0.5 or x >= N - 0.5:
        return 0.0

    # round position for pixel access
    cdef index_t i1 = my_round_index_t(x), i2

    # distance to pixel centre for weight calculation
    cdef double delta = i1 - x

    # choose one of the two neighbors to be incorporated
    # into interpolation
    if delta > 0:
        i2 = i1 - 1
    else:
        i2 = i1 + 1

    # perform mirroring for border pixels
    if i2 < 0:
        i2 = 1
    elif i2 >= N:
        i2 = N - 2

    # calculate weight
    cdef double w = w_linear(delta)

    # return interpolation result (weighted sum of grey values)
    return w * a[i1] + (1 - w) * a[i2]


cpdef double mirrored_interpolate_1d_py(np.float64_t[:] a, double x) nogil:
    """ Calculate interpolated grey values in a 1D float array, Python wrapper.

        Args:
            a: 1D numpy float64 array
            x: position at which the gray value should be interpolated
        Returns:
            interpolated value
    """
    cdef double *a_arr = <double *>malloc(sizeof(double) * a.shape[0])
    if a_arr == NULL:
        abort()
    cdef index_t i
    for i in range(a.shape[0]):
        a_arr[i] = <double> a[i]
    result = mirrored_interpolate_1d(a_arr, a.shape[0], x)
    free(a_arr)
    return result

# Typed memory views are used so that the GIL can be released
cdef double mirrored_interpolate_3d(np.int16_t[:, :, :] a, \
                                     double *x) nogil:
    """
        Perform 3D grey value interpolation.

        Note that the access order of the array is a[z, y, x] whereas
        x is a vector (x, y, z).

        Args:
            a: 3D array of int16 grey values, a[z, y, x]
            x: Position (x, y, z) at which grey value should be interpolated
        Returns:
            interpolated grey value (double)

    """

    # return 0 for positions outside of the image
    if not is_within_image(x, a):
        return 0.0

    cdef double delta

    # matrices containing weights and start indices
    # first index corresponds to the position (voxel 0 and 1)
    # second index corresponds to the axis (x, y, z)
    cdef double w[2][3]
    cdef index_t ind[2][3]

    cdef unsigned char i

    # i==0 corresponds to x, i==2 corresponds to y etc.
    for i in range(3):
        # voxel index for interpolation
        ind[0][i] = my_round_index_t(x[i])
        # argument for weight calculation
        delta = ind[0][i] -x[i]
        # index of second voxel to be used for interpolation
        ind[1][i] = ind[0][i] - 1 if delta > 0 else ind[0][i] + 1
        # perform mirroring
        if ind[1][i] < 0:
            ind[1][i] = 1
        elif ind[1][i] >= a.shape[2-i]:
            ind[1][i] = a.shape[2-i] - 2
        # weight value for fist voxel
        w[0][i] = w_linear(delta)
        # weight for second grey value
        w[1][i] = 1.0 - w[0][i]

    cdef double gval = 0.0
    cdef np.int16_t g
    cdef unsigned char i_x, i_y, i_z

    # iterate over 8 voxels to be considered
    for i_z in range(2):
        for i_y in range(2):
            for i_x in range(2):
                # get image grey value
                g = a[ind[i_z][2], ind[i_y][1], ind[i_x][0]]
                gval = gval+ g * w[i_z][2] * w[i_y][1] * w[i_x][0]
    return gval

cpdef double mirrored_interpolate_3d_py(np.int16_t[:, :, :] a,
        np.float64_t[:] x_np) nogil:
    """
        Python callable wrapper for 3D grey value interpolation.

        Note that the access order of the array is a[z, y, x] whereas
        x is a vector (x, y, z).

        Image Masking is represented implicitly by grey values as follows:
            grey values > 0: Voxels for which the similarity measure is to be
                calculated.
            grey values < 0: Voxels required for interpolation and similarity
                measure calculation in the vicinity of a voxel with
                grey value > 0
            grey value == 0: Background. Interpolation returns 0 for background
                voxels or positions close to the background.

        Args:
            a: 3D array of int16 grey values, a[z, y, x]
            x_np: Position (x, y, z) at which grey value should be interpolated
        Returns:
            interpolated grey value (double)
    """
    cdef double *x = <double *>malloc(3 * sizeof(double))
    if x == NULL:
        abort()

    x[0] = x_np[0]
    x[1] = x_np[1]
    x[2] = x_np[2]
    cdef double result = mirrored_interpolate_3d(a, x)
    free(x)
    return result

# Typed memory views are used so that the GIL can be released
cdef double masked_interpolate_3d(np.int16_t[:, :, :] a, \
                                     double *x) nogil:
    """
        Perform 3D grey value interpolation.

        Note that the access order of the array is a[z, y, x] whereas
        x is a vector (x, y, z).

        Image Masking is represented implicitly by grey values as follows:
            grey values > 0: Voxels for which the similarity measure is to be
                calculated.
            grey values < 0: Voxels required for interpolation and similarity
                measure calculation in the vicinity of a voxel with
                grey value > 0
            grey value == 0: Background. Interpolation returns 0 for background
                voxels or positions close to the background.

        Args:
            a: 3D array of int16 grey values, a[z, y, x]
            x: Position (x, y, z) at which grey value should be interpolated
        Returns:
            interpolated grey value (double)

    """

    # return 0 for positions outside of the image
    if not is_within_image_border_1(x, a):
        return 0.0

    cdef double delta

    # matrices containing weights and start indices
    # first index corresponds to the position (voxel 0 and 1)
    # second index corresponds to the axis (x, y, z)
    cdef double w[2][3]
    cdef index_t ind[2][3]

    cdef unsigned char i

    # i==0 corresponds to x, i==2 corresponds to y etc.
    for i in range(3):
        # voxel index for interpolation
        ind[0][i] = my_round_index_t(x[i])
        # argument for weight calculation
        delta = ind[0][i] - x[i]
        # weight value for fist voxel
        w[0][i] = w_linear(delta)
        # index of second voxel to be used for interpolation
        ind[1][i] = ind[0][i] - 1 if delta > 0 else ind[0][i] + 1
        # weight for second grey value
        w[1][i] = 1.0 - w[0][i]

    cdef double gval = 0.0, sign = -1
    cdef np.int16_t g
    cdef unsigned char i_x, i_y, i_z

    # iterate over 8 voxels to be considered
    for i_z in range(2):
        for i_y in range(2):
            for i_x in range(2):
                # get image grey value, return 0 if zero grey value is
                # encountered
                g = a[ind[i_z][2], ind[i_y][1], ind[i_x][0]]
                # check for masks and run length encoding
                if g == 0:
                    return 0.0
                # if one voxel in the environment posesses a positive
                # grey value, the result is positive to indicate that
                # this position is to be used for similarity measure
                # calculation
                if g > 0:
                    sign = 1
                else:
                    # use positive values for interpolation
                    g = -g
                # add contribution of this neighbor to the interpolation result
                gval = gval + g * w[i_z][2] * w[i_y][1] * w[i_x][0]

    return sign * gval


cpdef double masked_interpolate_3d_py(np.int16_t[:, :, :] a,
        np.float64_t[:] x_np) nogil:
    """
        Python callable wrapper for 3D grey value interpolation.

        Note that the access order of the array is a[z, y, x] whereas
        x is a vector (x, y, z).

        Image Masking is represented implicitly by grey values as follows:
            grey values > 0: Voxels for which the similarity measure is to be
                calculated.
            grey values < 0: Voxels required for interpolation and similarity
                measure calculation in the vicinity of a voxel with
                grey value > 0
            grey value == 0: Background. Interpolation returns 0 for background
                voxels or positions close to the background.

        Args:
            a: 3D array of int16 grey values, a[z, y, x]
            x_np: Position (x, y, z) at which grey value should be interpolated
        Returns:
            interpolated grey value (double)
    """
    cdef double *x = <double *>malloc(3 * sizeof(double))
    if x == NULL:
        abort()

    x[0] = x_np[0]
    x[1] = x_np[1]
    x[2] = x_np[2]
    cdef double result = masked_interpolate_3d(a, x)
    free(x)
    return result

@cython.cdivision(True)
cpdef void rescale_gvals(np.float64_t[:,:,:] in_im, 
        np.int16_t[:,:,:] out_im, double g_min, double g_max,
        np.int16_t out_min, np.int16_t out_max, np.float64_t epsilon,
        bint is_mask):
    cdef index_t x, y, z
    cdef np.int16_t result
    cdef np.float64_t gval_float64
    for z in range(in_im.shape[0]):
        for y in range(in_im.shape[1]):
            for x in range(in_im.shape[2]):
                result = 0
                gval_float64 = in_im[z, y, x]
                if is_mask: 
                    if gval_float64 != 0:
                        result = 1
                else:
                    if gval_float64 <= g_min:
                        result = out_min
                    elif gval_float64 >= g_max:
                        result = out_max
                    else:
                        result = <np.int16_t>((gval_float64 - g_min)
                                              / (g_max - g_min)
                                              * (out_max - out_min)
                                              + out_min + 0.5 + epsilon)
                out_im[z, y, x] = result


@cython.cdivision(True)
cpdef void downsample_XY(np.int16_t[:,:,:] in_im, np.float64_t[:, :, :] out_im,
                  np.float64_t[:] scale_factor,
                  np.float64_t[:] offset, int thread_num = -1) nogil:

    """
        Downsampling of an image in the xy-plane.

        in_im is resampled to the shape of out_im. First, resampling in x
        direction is done, the results are then resampled in y direction.
        The algorithm is based on 1D "integral images" similar to the method
        used in the SURF algorithm. The method allows for arbitrary scaling
        factors. However only downsampling is supported so that the size of
        out_im must not exceed the size of in_im in any direction.

        Args:
            in_im (memoryview, np.int16): input image
            out_im (memoryview, np.float64): output image
            scale_factor (numpy array): 3D scaling factor > 1 for downsampling
            thread_num (int): number of threads to use. Values < 1 indicate
                that all CPU threads should be used.
    """

    # check image sizes
    if in_im.shape[0] > out_im.shape[0]:
        abort()

    # only downsampling is supported
    if in_im.shape[1] < out_im.shape[1] or in_im.shape[2] < out_im.shape[2]:
        abort()

    if thread_num < 1:
        thread_num = openmp.omp_get_max_threads()

    cdef index_t z, y, x
    cdef double *im_slice
    cdef double *im_row_y
    cdef double *im_row_x
    cdef double sum_left, sum_right
    cdef double scale_factor_x = scale_factor[0]
    cdef double scale_factor_y = scale_factor[1]

    with nogil, parallel(num_threads=thread_num):
        im_slice = <double *>malloc(sizeof(double) * \
                    in_im.shape[1] * out_im.shape[2])
        if im_slice == NULL:
            abort()
        im_row_x   = <double *>malloc(sizeof(double) * in_im.shape[2])
        if im_row_x == NULL:
            abort()
        im_row_y   = <double *>malloc(sizeof(double) * in_im.shape[1])
        if im_row_y == NULL:
            abort()
        # work z-slice by z-slice
        for z in prange(in_im.shape[0], schedule='dynamic'):
            # perform row by row x resampling and store result in im_slice
            for y in range(in_im.shape[1]):
                # accumulate image row into buffer
                im_row_x[0] = in_im[z, y, 0]
                for x in range(1, in_im.shape[2]):
                    im_row_x[x] = im_row_x[x-1] + in_im[z, y, x]
                # downsample in x direction into im_slice
                sum_left = mirrored_interpolate_1d(im_row_x,
                            in_im.shape[2], offset[0] - 1.0)
                for x in range(out_im.shape[2]):
                    sum_right = mirrored_interpolate_1d(im_row_x,
                            in_im.shape[2],
                            (x + 1.0) * scale_factor_x - 1.0 + offset[0])
                    im_slice[x + y * out_im.shape[2]] =\
                        (sum_right - sum_left) / scale_factor_x
                    sum_left = sum_right
            # perform colum by column y resampling and store result in out_im
            for x in range(out_im.shape[2]):
                im_row_y[0] = im_slice[x]
                #accumulate into im_row_y
                for y in range(1, in_im.shape[1]):
                    im_row_y[y] = im_row_y[y-1] + \
                            im_slice[x + y * out_im.shape[2]]
                # downsample into out_im
                sum_left = mirrored_interpolate_1d(im_row_y,\
                            in_im.shape[1], offset[1] - 1.0)
                for y in range(out_im.shape[1]):
                    sum_right = mirrored_interpolate_1d(im_row_y,\
                            in_im.shape[1],
                            (y + 1.0) * scale_factor_y - 1.0 + offset[1])
                    out_im[z, y, x] = \
                            <np.float64_t>((sum_right - sum_left) /\
                                scale_factor_y)
                    sum_left = sum_right

        free(im_slice)
        free(im_row_x)
        free(im_row_y)

cpdef void downsample_Z(np.float64_t[:,:,:] in_im,
                        np.int16_t[:, :, :] out_im,
                        np.float64_t[:] scale_factor,
                        np.float64_t[:] offset,
                        int thread_num = -1, 
                        bint is_mask = False) nogil:
    """
        Downsampling of an image in the z direction.

        in_im is resampled to the shape of out_im. Grey values results are
        resampled in z direction.
        The algorithm is based on 1D "integral images" similar to the method
        used in the SURF algorithm. The method allows for arbitrary scaling
        factors. However only downsampling is supported so that the size of
        out_im must not exceed the size of in_im in any direction.

        Args:
            in_im (memoryview, np.float64):  input image
            out_im (memoryview, np.int16): output image
            scale_factor (numpy array): 3D Scaling factor > 1 for downsampling
            offset (memoryview, np.float64): offset to be added to source
                                             image coordinates
            thread_num (int): number of threads to use. Values < 1 indicate
                that all CPU threads should be used.
            is_mask (bool): True if image should be treated as a binary mask
    """

    # check image sizes
    if in_im.shape[1] != out_im.shape[1] or in_im.shape[2] != out_im.shape[2]:
        abort()

    # only downsampling is supported
    if in_im.shape[0] < out_im.shape[0]:
        abort()

    if thread_num < 1:
        thread_num = openmp.omp_get_max_threads()

    cdef index_t x, y, z
    cdef double *im_row_z
    cdef double scale_factor_z = scale_factor[2]
    cdef double sum_left, sum_right, g_val

    with nogil, parallel(num_threads=thread_num):
        im_row_z   = <double *>malloc(sizeof(double) * in_im.shape[0])
        if im_row_z == NULL:
            abort()
        # work y-slice by y-slice
        for y in prange(in_im.shape[1], schedule='dynamic'):
            # perform stack by stack resampling and store result in im_row_z
            for x in range(in_im.shape[2]):
                # accumulate into im_row_z
                im_row_z[0]=in_im[0, y, x]
                for z in range(1, in_im.shape[0]):
                    im_row_z[z] = im_row_z[z-1] + in_im[z, y, x]
                # downsample into out_im
                sum_left = mirrored_interpolate_1d(im_row_z,
                            in_im.shape[0], offset[2] - 1.0)
                for z in range(out_im.shape[0]):
                    sum_right = mirrored_interpolate_1d(im_row_z,
                            in_im.shape[0],
                            (z + 1.0) * scale_factor_z - 1.0 + offset[2])
                    g_val = (sum_right - sum_left) / scale_factor_z
                    if not is_mask:
                        out_im[z, y, x] = my_round_int16(g_val)
                    else:
                        if g_val > 0.25:
                            out_im[z, y, x] = <np.int16_t>1
                        else:
                            out_im[z, y, x] = <np.int16_t>0
                    sum_left = sum_right
        free(im_row_z)

cpdef void downsample(np.int16_t[:,:,:] in_im,
                      np.int16_t[:, :, :] out_im,
                      int thread_num = -1,
                      bint is_mask = False):
    """
        Downsampling of a 3D image.

        in_im is resampled to the shape of out_im.
        The algorithm is based on 1D "integral images" similar to the method
        used in the SURF algorithm. The method allows for arbitrary scaling
        factors. However only downsampling is supported so that the size of
        out_im must not exceed the size of in_im in any direction. Resampling
        is performed in two steps. First, in-slice resampling (x and y) of
        in_im is performed into an intermediate float image which is
        afterwards resampled in z direction into out_im.

        Args:
            in_im:  input image  (memoryview, np.int16)
            out_im: output image (memoryview, np.int16)
            thread_num: number of threads to use. Values < 1 indicate
                that all CPU threads should be used.
            is_mask (bool): True if image should be treated as a binary mask
    """
    cdef np.float64_t[:,:,:] tmp_im = np.zeros((in_im.shape[0], 
        out_im.shape[1], out_im.shape[2]), dtype=np.float64)
    cdef np.float64_t[:] offset = np.zeros(3, dtype=np.float64)
    cdef np.float64_t[:] scale_factor = np.zeros(3, dtype=np.float64)
    cdef int i
    for i in range(3):
        scale_factor[i] = <double>in_im.shape[2-i] / <double>out_im.shape[2-i]
    downsample_XY(in_im, tmp_im, scale_factor, offset, thread_num)
    downsample_Z(tmp_im, out_im, scale_factor, offset, thread_num, is_mask)
    del(tmp_im)


cpdef void downsample_with_offset(np.int16_t[:,:,:] in_im,
                      np.int16_t[:, :, :] out_im,
                      np.float64_t[:, :, :] tmp_im,
                      np.float64_t[:] scale_factor,
                      np.float64_t[:] offset,
                      int thread_num = -1,
                      bint is_mask = False):
    """
        Downsampling of a 3D image part.

        in_im is resampled to the shape of out_im.
        The algorithm is based on 1D "integral images" similar to the method
        used in the SURF algorithm. The method allows for arbitrary scaling
        factors. However only downsampling is supported so that the size of
        out_im must not exceed the size of in_im in any direction. Resampling
        is performed in two steps. First, in-slice resampling (x and y) of
        in_im is performed into an intermediate float image which is
        afterwards resampled in z direction into out_im.

        Args:
            in_im (memoryview, np.int16):  input image
            out_im (memoryview, np.int16): output image
            tmp_im (memoryview, np.float64): temporary image holding in-plane
                                            downsampling results
            scale_factor (numpy array): 3D scaling factor > 1 for downsampling
            offset (memoryview, np.float64): offset to be added to source
                                             image coordinates
            thread_num (int): number of threads to use. Values < 1 indicate
                that all CPU threads should be used.
            is_mask bool): True if image should be treated as a binary mask
    """

    downsample_XY(in_im, tmp_im, scale_factor, offset, thread_num)
    downsample_Z(tmp_im, out_im, scale_factor, offset, thread_num, is_mask)

cpdef void interpolate_block(np.int16_t[:,:,:] fixed_im,
                      np.int16_t[:, :, :] moving_im,
                      np.float64_t[:, :] t_matrix_fixed_to_moving,
                      int thread_num = -1,
                      bint is_mask = False):

    if thread_num < 1:
        thread_num = openmp.omp_get_max_threads()

    cdef index_t x, y, z
    cdef double interpolation_result
    cdef np.int16_t out_grey_val
    with nogil, parallel(num_threads=thread_num):
        p_fixed = <double *>malloc(3 * sizeof(double))
        p_moving = <double *>malloc(3 * sizeof(double))
        for z in prange(fixed_im.shape[0], schedule='dynamic'):
            p_fixed[2] = <double>z
            for y in range(fixed_im.shape[1]):
                p_fixed[1] = <double>y
                for x in range(fixed_im.shape[2]):
                    p_fixed[0] = <double>x
                    transform_point(t_matrix_fixed_to_moving, 
                                    p_fixed, p_moving)
                    interpolation_result = \
                            mirrored_interpolate_3d(moving_im, p_moving)
                    if is_mask:
                        fixed_im[z, y, x] =\
                            1 if interpolation_result > 0.4 else 0
                    else:
                        fixed_im[z, y, x] = \
                                my_round_int16(interpolation_result)
        free(p_fixed)
        free(p_moving)


cpdef void fill_gm_image(np.int16_t[:,:,:] in_block, 
        np.int16_t[:,:,:] mask_block, np.float32_t[:,:,:] gm_block, 
        int thread_num):

    cdef np.float64_t d_x, d_y, d_z
    cdef index_t x, y, z, z_in
    for z in prange(gm_block.shape[0], nogil=True, 
                schedule='dynamic', num_threads=thread_num):
        z_in = z + 1
        for y in range(1, gm_block.shape[1]-1):
            for x in range(1, gm_block.shape[2]-1):
                if mask_block is None or mask_block[z, y, x] != 0:
                    d_x = in_block[z_in, y, x+1] - in_block[z_in, y, x-1]
                    d_y = in_block[z_in, y+1, x] - in_block[z_in, y-1, x]
                    d_z = in_block[z_in+1, y, x] - in_block[z_in-1, y, x]
                    gm_block[z, y, x] = \
                        <np.float32_t> d_x * d_x + d_y * d_y + d_z * d_z

cpdef np.float32_t gm_max(np.float32_t[:,:,:] gm_block): 
    cdef index_t x, y, z
    cdef np.float32_t gm_max = -1
    for z in range(gm_block.shape[0]):
        for y in range(1, gm_block.shape[1]-1):
            for x in range(1, gm_block.shape[2]-1):
                gm_max = max(gm_max, gm_block[z, y, x])
    return gm_max

cpdef void fill_gm_histogram(np.float32_t[:,:,:] gm_block,
        np.uint64_t[:] hist, double gm_max):

    cdef np.float64_t gm
    cdef index_t x, y, z
    cdef size_t bin_nr, n_bins = hist.shape[0]
    for z in range(gm_block.shape[0]):
        for y in range(1, gm_block.shape[1]-1):
            for x in range(1, gm_block.shape[2]-1):
                gm = gm_block[z, y, x]
                bin_nr = my_round_size_t((gm / gm_max) * n_bins)
                bin_nr = min(bin_nr, n_bins-1)
                hist[bin_nr] += 1


cpdef void gm_masking(np.int16_t[:,:,:] in_im, 
        np.int16_t[:,:,:] out_im, np.float32_t[:,:,:] gm_im,
        int z_offset, double gm_limit, 
        bint is_moving): 

    cdef index_t x, y, z, z_out
    
    for z in range(gm_im.shape[0]):
        z_out = z + z_offset
        for y in range(1, gm_im.shape[1]-1):
            for x in range(1, gm_im.shape[2]-1):
                if gm_im[z, y, x] > gm_limit:
                    out_im[z_out, y, x] = in_im[z_out, y, x]
                    set_gm_environment(in_im, out_im, 
                        x, y, z_out, is_moving)


cpdef void do_run_length_encoding(np.int16_t[:,:,:] im, int thread_num) nogil:
    cdef size_t slice_pos=0, rle_start=0, rle_len
    cdef size_t pixels_per_slice = im.shape[1] * im.shape[2]
    cdef bint rle_active=False

    cdef index_t z

    for z in prange(im.shape[0], nogil=True, 
                schedule='dynamic', num_threads=thread_num):
        rle_active = False
        rle_len = 0
        for slice_pos in range(pixels_per_slice-1):
            # switch on rle for two consecutive zero output voxels 
            if not rle_active and \
                    im[z, slice_pos // im.shape[2], 
                      slice_pos % im.shape[2]] == 0 and \
                    im[z, (slice_pos + 1) // im.shape[2], 
                      (slice_pos + 1) % im.shape[2]] == 0 and \
                    slice_pos < pixels_per_slice - 2:
                rle_start = slice_pos
                rle_active = True
                rle_len = 0
            # count rle pixels        
            if rle_active and im[z, slice_pos // im.shape[2],
                                     slice_pos % im.shape[2]] <= 0:
                rle_len = rle_len + 1  
            # store rle result for next positive grey value
            # or if the end of the slice has been reached
            if rle_active and \
                    (im[z, slice_pos // im.shape[2], 
                            slice_pos % im.shape[2]] > 0 or \
                     slice_pos >= pixels_per_slice - 2):
                rle_active = False
                # assume cache line size of 64 and restrict rle
                # usage to certain cache misses
                if rle_len > 32 and rle_len < 2**15 * (2**15 - 2**14):
                    im[z, rle_start // im.shape[2], 
                           rle_start % im.shape[2]] = \
                           <np.int16_t> -(rle_len // 2**15  + 2**14) 
                    im[z, (rle_start + 1) // im.shape[2], 
                           (rle_start+1) % im.shape[2]] = \
                           <np.int16_t> -(rle_len % 2**15)


cpdef void remove_rle(np.int16_t[:, :] in_slice, np.int16_t[:, :] out_slice):
    cdef index_t x, y, x_size = in_slice.shape[1]
    cdef size_t slice_pos=0 
    cdef size_t nr_of_pixels = in_slice.shape[0] * in_slice.shape[1]
    
    while slice_pos < nr_of_pixels:
        y = slice_pos // x_size
        x = slice_pos % x_size
        if in_slice[y, x] > -2**14:
            out_slice[y, x] = in_slice[y, x]
            slice_pos = slice_pos + 1
        else:
            out_slice[y, x] = 0
            out_slice[(slice_pos+1) // x_size, (slice_pos+1) % x_size] = 0
            slice_pos = slice_pos + 2
        


cpdef void add_inertia_tensor_slice(np.int16_t[:, :] im,
        index_t z, np.float64_t[:, :] voxel_to_world_matrix, 
        np.int64_t[:] N, np.float64_t[:] centre, np.float64_t[:, :] TI):
    
    cdef index_t x, y
    cdef double *p_v = <double *>malloc(sizeof(double) * 3)
    cdef double *p_w = <double *>malloc(sizeof(double) * 3)

    p_v[2] = z
    for y in range(1, im.shape[0]-1):
        p_v[1] = y
        for x in range(1, im.shape[1]-1):
            if im[y, x] > 0:
                p_v[0] = x
                transform_point(voxel_to_world_matrix, p_v, p_w)
                N[0] += 1
                centre[0] += p_w[0]
                centre[1] += p_w[1]
                centre[2] += p_w[2]
                TI[0, 0] += p_w[1] * p_w[1] + p_w[2] * p_w[2]
                TI[0, 1] -= p_w[0] * p_w[1]
                TI[0, 2] -= p_w[0] * p_w[2]
                TI[1, 1] += p_w[0] * p_w[0] + p_w[2] * p_w[2]
                TI[1, 2] -= p_w[1] * p_w[2]
                TI[2, 2] += p_w[0] * p_w[0] + p_w[1] * p_w[1]
    free(p_v)
    free(p_w)

cdef void lcc_process_environment(
        np.int16_t[:, :, :] fixed_im, 
        np.int16_t[:, :, :] moving_im,
        np.float64_t[:, :] trans_mat_vox, 
        np.int64_t[:] N_buf,
        np.float64_t[:] lcc_buf, 
        index_t x_f, index_t y_f, index_t z_f, 
        double *p_fixed, double *p_t,
        np.int16_t *b_buf, np.float64_t *t_buf,
        np.int16_t b_val_c, np.float64_t t_val_c) nogil:

    cdef double b, t, b_avg, t_avg, b_i_t_i, b_i_2, t_i_2, denom
    cdef bint pixel_outside
    cdef index_t i, x_l, y_l, z_l, d_z, d_z_y

    i=0
    b_avg = 0.0
    t_avg = 0.0
    pixel_outside = False
    for z_l in range(-1, 2):
        if pixel_outside:
            break
        d_z = my_abs(z_l)
        for y_l in range(-1+d_z, 2-d_z):
            d_z_y = d_z + my_abs(y_l)
            for x_l in range(-1+d_z_y, 2-d_z_y):
                if d_z_y == 0 and x_l == 0:
                    b_buf[i] = b_val_c
                    t_buf[i] = t_val_c
                else:
                    b_buf[i] = my_abs(fixed_im[z_f+z_l, y_f+y_l, x_f+x_l])
                    p_fixed[0] = x_f + x_l
                    p_fixed[1] = y_f + y_l 
                    p_fixed[2] = z_f + z_l
                    transform_point(trans_mat_vox, p_fixed, p_t)
                    t_buf[i] = my_fabs(masked_interpolate_3d( moving_im, p_t))
                if t_buf[i] < 1 or b_buf[i] < 1:
                    pixel_outside = True
                else:
                    b_avg = b_avg + <double> b_buf[i]
                    t_avg = t_avg + t_buf[i]
                    i = i + 1
    if not pixel_outside:
        b_avg = b_avg / 7
        t_avg = t_avg / 7
        b_i_t_i = 0
        b_i_2 = 0
        t_i_2 = 0
        for i in range(7):
            b = <double> b_buf[i] - b_avg
            t = t_buf[i] - t_avg
            b_i_t_i = b_i_t_i + b * t
            b_i_2 = b_i_2 + b * b
            t_i_2 = t_i_2 + t * t
        denom = b_i_2 * t_i_2
        if denom > 1e-20:
            N_buf[z_f] = N_buf[z_f] + 1
            lcc_buf[z_f] = lcc_buf[z_f] +\
                    1.0 - (b_i_t_i * b_i_t_i) / denom

cpdef void lcc_block(
        np.int16_t[:, :, :] fixed_im, 
        np.int16_t[:, :, :] moving_im,
        np.float64_t[:, :] trans_mat_vox, 
        np.int64_t[:] N_buf,
        np.float64_t[:] lcc_buf, 
        np.int32_t thread_num,
        bint use_rle=True):

    cdef index_t z_f, y_f, x_f, x_size = fixed_im.shape[2]
    cdef np.int16_t *b_buf
    cdef np.int16_t b_val_c
    cdef np.float64_t *t_buf
    cdef np.float64_t t_val_c
    cdef double *p_fixed
    cdef double *p_t
    cdef size_t slice_index
    cdef size_t pixels_in_slice = fixed_im.shape[1] * fixed_im.shape[2]

    if thread_num < 1:
        thread_num = min(openmp.omp_get_max_threads(), fixed_im.shape[1]-4)

    with nogil, parallel(num_threads=thread_num):
        b_buf = <np.int16_t *>malloc(7 * sizeof(np.int16_t))
        t_buf = <np.float64_t *>malloc(7 * sizeof(np.float64_t))
        p_fixed = <double *>malloc(3 * sizeof(double))
        p_t = <double *>malloc(3 * sizeof(double))

        for z_f in prange(2, fixed_im.shape[0]-2, schedule='dynamic'):
            if use_rle:
                slice_index = 0
                while slice_index < pixels_in_slice:
                    x_f = slice_index % x_size
                    y_f = slice_index // x_size
                    b_val_c = fixed_im[z_f, y_f, x_f] 
                    if  b_val_c > 0 and \
                            x_f > 1 and x_f < fixed_im.shape[2] - 2 and \
                            y_f > 1 and y_f < fixed_im.shape[1] - 2:
                        p_fixed[2] = <double>z_f
                        p_fixed[1] = <double>y_f
                        p_fixed[0] = <double>x_f
                        transform_point(trans_mat_vox, p_fixed, p_t)
                        t_val_c = masked_interpolate_3d(moving_im, p_t)
                        if t_val_c > 0:
                            lcc_process_environment(fixed_im, moving_im,
                                trans_mat_vox, N_buf, lcc_buf, 
                                x_f, y_f, z_f, p_fixed, p_t,
                                b_buf, t_buf, b_val_c, t_val_c)
                    if b_val_c <= -2**14 and \
                            slice_index < pixels_in_slice - 1: 
                        slice_index = slice_index +\
                                      (-b_val_c - 2**14) * 2**15 -\
                                      fixed_im[z_f, 
                                               (slice_index + 1) // x_size,
                                               (slice_index + 1) % x_size]
                    else:
                        slice_index = slice_index + 1
            else:
                for y_f in range(2, fixed_im.shape[1]-2):
                    for x_f in range(2, fixed_im.shape[2]-2):
                        b_val_c = fixed_im[z_f, y_f, x_f] 
                        if  b_val_c > 0:
                            p_fixed[2] = <double>z_f
                            p_fixed[1] = <double>y_f
                            p_fixed[0] = <double>x_f
                            transform_point(trans_mat_vox, p_fixed, p_t)
                            t_val_c = masked_interpolate_3d(moving_im, p_t)
                            if t_val_c > 0:
                                lcc_process_environment(fixed_im, moving_im,
                                    trans_mat_vox, N_buf, lcc_buf, 
                                    x_f, y_f, z_f, p_fixed, p_t,
                                    b_buf, t_buf, b_val_c, t_val_c)
        free(b_buf)
        free(t_buf)
        free(p_fixed)
        free(p_t)

cdef void lcc_process_environment_with_derivatives(
        np.int16_t[:, :, :] fixed_im, 
        np.int16_t[:, :, :] moving_im,
        np.float64_t[:, :] fixed_v2w_matrix,
        np.float64_t[:, :] trans_mat_vox, 
        np.float64_t[:, :] trans_mat_gradient, 
        np.int32_t n_p,
        np.int64_t[:] N_buf,
        np.float64_t[:] lcc_buf, 
        np.float64_t[:, :] gradient_buf, 
        np.float64_t[:, :, :] hessian_buf, 
        index_t x_f, index_t y_f, index_t z_f, 
        double *p_fixed, double *p_t,
        np.int16_t *t_buf, np.float64_t *b_buf,
        double *grad_v, double *grad_w,
        double *derivative_buf, 
        double *derivative_avg,
        double *t_t_der_sum,
        double *b_t_der_sum,
        double *m_ij_der,
        np.int16_t t_val_c, np.float64_t b_val_c) nogil:

    cdef index_t x_l, y_l, z_l, d_z, d_z_y, x_t, y_t, z_t, i, k, l
    cdef double b_2_sum, t_2_sum, b_t_sum, b_i, m_ij
    cdef double b, t, b_avg, t_avg, b_i_t_i, b_i_2, t_i_2, denom, c_j
    cdef bint pixel_outside
    cdef np.int16_t b_int

    # fill moving buffer (b_buf due to image swapping ...)
    for z_l in range(-1, 2):
        d_z = my_abs(z_l)
        for y_l in range(-1+d_z, 2-d_z):
            d_z_y = d_z + my_abs(y_l)
            for x_l in range(-1+d_z_y, 2-d_z_y):
                if d_z_y == 0 and x_l == 0:
                    t = b_val_c
                else:
                    p_fixed[0] = x_f + x_l
                    p_fixed[1] = y_f + y_l 
                    p_fixed[2] = z_f + z_l
                    transform_point(trans_mat_vox, p_fixed, p_t)
                    t = my_fabs(masked_interpolate_3d( moving_im, p_t))
                if t < 1:
                    return
                set_3d_array_3(b_buf, z_l+1, y_l+1, x_l+1, t)

    # fill fixed buffer (t_buf due to image swapping ...)
    for z_l in range(-2, 3):
        d_z = my_abs(z_l)
        for y_l in range(-2+d_z, 3-d_z):
            d_z_y = d_z + my_abs(y_l)
            for x_l in range(-2+d_z_y, 3-d_z_y):
                if d_z_y == 0 and x_l == 0:
                    b_int = t_val_c
                else:
                    b_int = my_abs(fixed_im[z_f+z_l, y_f+y_l, x_f+x_l])
                if b_int < 1:
                    return
                else:
                    set_3d_array_5(t_buf, z_l+2, y_l+2, x_l+2, b_int)

    # add up values and calculate gradients
    i=0
    t_avg = 0.0
    b_avg = 0.0
    for z_l in range(-1, 2):
        d_z = my_abs(z_l)
        z_t = z_l + 2
        p_fixed[2] = <double>(z_f + z_l)
        for y_l in range(-1+d_z, 2-d_z):
            d_z_y = d_z + my_abs(y_l)
            y_t = y_l + 2
            p_fixed[1] = <double>(y_f + y_l)
            for x_l in range(-1+d_z_y, 2-d_z_y):
                x_t = x_l + 2
                b_avg = b_avg + get_3d_array_3(b_buf, z_l+1, y_l+1, x_l+1)
                t_avg = t_avg + <double> get_3d_array_5(t_buf, z_t, y_t, x_t)
                # gradient in voxel coordinates
                grad_v[0] = <double>(
                                get_3d_array_5(t_buf, z_t, y_t, x_t+1) -\
                                get_3d_array_5(t_buf, z_t, y_t, x_t-1))
                grad_v[1] = <double>(
                                get_3d_array_5(t_buf, z_t, y_t+1, x_t) -\
                                get_3d_array_5(t_buf, z_t, y_t-1, x_t))
                grad_v[2] = <double>(
                                get_3d_array_5(t_buf, z_t+1, y_t, x_t) -\
                                get_3d_array_5(t_buf, z_t-1, y_t, x_t))
                # convert into world coordinates
                transform_vector(trans_mat_gradient, grad_v, grad_w)
                # transform voxel position into world coordinates
                p_fixed[0] = <double>x_f + x_l
                transform_point(fixed_v2w_matrix, p_fixed, p_t)
                # calculate derivatives
                if n_p == 6:
                    # translation
                    for l in range(3):
                        derivative_buf[i * n_p + l] = grad_w[l]
                    # rotation
                        derivative_buf[i * n_p + 3] = \
                            grad_w[2]*p_t[1] - grad_w[1]*p_t[2]
                        derivative_buf[i * n_p + 4] = \
                            grad_w[0]*p_t[2] - grad_w[2]*p_t[0]
                        derivative_buf[i * n_p + 5] = \
                            grad_w[1]*p_t[0] - grad_w[0]*p_t[1]
                else:
                    # homogenous coordinates
                    p_t[3] = 1.0
                    for k in range(n_p):
                        derivative_buf[i * n_p + k] = grad_w[k/4] * p_t[k%4]
                i = i + 1

    # calculate average grey values
    b_avg = b_avg / 7.0
    t_avg = t_avg / 7.0

    # calculate average derivative values
    for k in range(n_p):
        derivative_avg[k] = 0.0
        for i in range(7):
            derivative_avg[k] = derivative_avg[k] +  derivative_buf[i*n_p+k] 
        derivative_avg[k] = derivative_avg[k] / 7.0

    # calculate sums
    b_2_sum = 0.0
    t_2_sum = 0.0
    b_t_sum = 0.0
    for i in range(n_p):
       t_t_der_sum[i] = 0.0
       b_t_der_sum[i] = 0.0
    i=0
    for z_l in range(-1, 2):
        d_z = my_abs(z_l)
        for y_l in range(-1+d_z, 2-d_z):
            d_z_y = d_z + my_abs(y_l)
            for x_l in range(-1+d_z_y, 2-d_z_y):
                b_i = get_3d_array_3(b_buf, z_l+1, y_l+1, x_l+1) - b_avg
                t_i = <double>get_3d_array_5(t_buf, z_l+2, y_l+2, x_l+2) - t_avg
                b_2_sum = b_2_sum + b_i * b_i
                t_2_sum = t_2_sum + t_i * t_i
                b_t_sum = b_t_sum + b_i * t_i
                for k in range(n_p):
                    t_t_der_sum[k] = t_t_der_sum[k] +\
                        t_i * (derivative_buf[i*n_p+k] - derivative_avg[k])
                    b_t_der_sum[k] = + b_t_der_sum[k] +\
                        b_i * (derivative_buf[i*n_p+k] - derivative_avg[k])
                i = i + 1
        
    if b_2_sum < 1e-20 or t_2_sum < 1e-20:
        return
            
    # update lcc, gradient and hessian buffers
    i=0
    for z_l in range(-1, 2):
        d_z = my_abs(z_l)
        for y_l in range(-1+d_z, 2-d_z):
            d_z_y = d_z + my_abs(y_l)
            for x_l in range(-1+d_z_y, 2-d_z_y):
                c_j = 2.0 / b_2_sum
                b_i = get_3d_array_3(b_buf, z_l+1, y_l+1, x_l+1) - b_avg
                t_i = <double>get_3d_array_5(t_buf, z_l+2, y_l+2, x_l+2) - t_avg
                m_ij = b_i - b_t_sum / t_2_sum * t_i
                lcc_buf[z_f] = lcc_buf[z_f] + c_j * m_ij * m_ij
                for k in range(n_p):
                    m_ij_der[k] = 2 * b_t_sum * t_t_der_sum[k] * t_i / \
                                    (t_2_sum * t_2_sum) -\
                        (b_t_der_sum[k] * t_i + b_t_sum *\
                            (derivative_buf[i*n_p+k] - derivative_avg[k])) / \
                        t_2_sum
                for k in range(n_p):
                    gradient_buf[z_f, k] = gradient_buf[z_f, k] + \
                                           c_j * m_ij * m_ij_der[k]
                    for l in range(k, n_p):
                        hessian_buf[z_f, k, l] = hessian_buf[z_f, k, l] + \
                            c_j * m_ij_der[k] * m_ij_der[l]
                i = i + 1     
    N_buf[z_f] = N_buf[z_f] + 1

cpdef void lcc_block_with_derivatives(
        np.int16_t[:, :, :] fixed_im, 
        np.int16_t[:, :, :] moving_im,
        np.float64_t[:, :] fixed_v2w_matrix,
        np.float64_t[:, :] trans_mat_vox, 
        np.float64_t[:,:] trans_mat_gradient,
        np.int32_t n_p,
        np.int64_t[:] N_buf,
        np.float64_t[:] lcc_buf, 
        np.float64_t[:, :] gradient_buf,
        np.float64_t[:, :, :] hessian_buf,
        np.int32_t thread_num,
        bint use_rle=True):

    cdef index_t z_f, y_f, x_f, x_size = fixed_im.shape[2]
    cdef np.int16_t *b_buf
    cdef np.float64_t *t_buf
    cdef double *p_fixed
    cdef double *p_t
    cdef double *grad_v
    cdef double *grad_w
    cdef double *derivative_buf
    cdef double *derivative_avg
    cdef double *t_t_der_sum
    cdef double *b_t_der_sum
    cdef bint pixel_outside
    cdef np.int16_t b_val_c
    cdef np.float64_t t_val_c
    cdef size_t slice_index
    cdef size_t pixels_in_slice = fixed_im.shape[1] * fixed_im.shape[2]

    if thread_num < 1:
        thread_num = min(openmp.omp_get_max_threads(), fixed_im.shape[1]-4)

    with nogil, parallel(num_threads=thread_num):
        b_buf = <np.int16_t *>malloc(5 * 5 * 5 * sizeof(np.int16_t))
        t_buf = <np.float64_t *>malloc(3 * 3 * 3 * sizeof(np.float64_t))
        p_fixed = <double *>malloc(3 * sizeof(double))
        p_t = <double *>malloc(4 * sizeof(double))
        grad_v = <double *>malloc(3 * sizeof(double))
        grad_w = <double *>malloc(3 * sizeof(double))
        derivative_buf = <double *>malloc(n_p * 7 * sizeof(double))
        derivative_avg = <double *>malloc(n_p * sizeof(double))
        t_t_der_sum = <double *>malloc(n_p * sizeof(double))
        b_t_der_sum = <double *>malloc(n_p * sizeof(double))
        m_ij_der = <double *>malloc(n_p * sizeof(double))

        for z_f in prange(2, fixed_im.shape[0]-2, schedule='dynamic'):
            if use_rle:
                slice_index = 0
                while slice_index < pixels_in_slice:
                    x_f = slice_index % x_size
                    y_f = slice_index // x_size
                    b_val_c = fixed_im[z_f, y_f, x_f] 
                    if  b_val_c > 0 and \
                            x_f > 1 and x_f < fixed_im.shape[2] - 2 and \
                            y_f > 1 and y_f < fixed_im.shape[1] - 2:
                        p_fixed[2] = <double>z_f
                        p_fixed[1] = <double>y_f
                        p_fixed[0] = <double>x_f
                        transform_point(trans_mat_vox, p_fixed, p_t)
                        t_val_c = masked_interpolate_3d(moving_im, p_t)
                        if t_val_c > 0:
                            lcc_process_environment_with_derivatives(
                                        fixed_im,  moving_im,
                                        fixed_v2w_matrix,
                                        trans_mat_vox, 
                                        trans_mat_gradient, 
                                        n_p,
                                        N_buf, lcc_buf, 
                                        gradient_buf, hessian_buf, 
                                        x_f, y_f, z_f, 
                                        p_fixed, p_t,
                                        b_buf, t_buf,
                                        grad_v, grad_w,
                                        derivative_buf, 
                                        derivative_avg,
                                        t_t_der_sum,
                                        b_t_der_sum,
                                        m_ij_der,
                                        b_val_c, t_val_c)
                    if b_val_c <= -2**14 and \
                            slice_index < pixels_in_slice - 1: 
                        slice_index = slice_index +\
                                      (-b_val_c - 2**14) * 2**15 -\
                                      fixed_im[z_f, 
                                               (slice_index + 1) // x_size,
                                               (slice_index + 1) % x_size]
                    else:
                        slice_index = slice_index + 1
            else:
                for y_f in range(2, fixed_im.shape[1]-2):
                    for x_f in range(2, fixed_im.shape[2]-2):
                        b_val_c = fixed_im[z_f, y_f, x_f] 
                        if b_val_c > 0:
                            p_fixed[2] = <double>z_f
                            p_fixed[1] = <double>y_f
                            p_fixed[0] = <double>x_f
                            transform_point(trans_mat_vox, p_fixed, p_t)
                            t_val_c = masked_interpolate_3d(moving_im, p_t)
                            if t_val_c > 0:
                                lcc_process_environment_with_derivatives(
                                        fixed_im,  moving_im,
                                        fixed_v2w_matrix,
                                        trans_mat_vox, 
                                        trans_mat_gradient, 
                                        n_p,
                                        N_buf, lcc_buf, 
                                        gradient_buf, hessian_buf, 
                                        x_f, y_f, z_f, 
                                        p_fixed, p_t,
                                        b_buf, t_buf,
                                        grad_v, grad_w,
                                        derivative_buf, 
                                        derivative_avg,
                                        t_t_der_sum,
                                        b_t_der_sum,
                                        m_ij_der,
                                        b_val_c, t_val_c)
        free(b_buf)
        free(t_buf)
        free(p_fixed)
        free(p_t)
        free(grad_v)
        free(grad_w)
        free(derivative_buf)
        free(derivative_avg)
        free(t_t_der_sum)
        free(b_t_der_sum)
        free(m_ij_der)



# lcreg: Efficent rigid and affine 3D image registration
# 
# Copyright (C) 2019  Peter Rösch, Peter.Roesch@hs-augsburg.de
# 
# Organisation:
# Faculty of Computer Science, Augsburg University of Applied Sciences,
# An der Hochschule 1, 86161 Augsburg, Germany
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

#cython: language_level=3, boundscheck=False, profile=False, cdivision=True, wraparound=False
cimport numpy as np
import numpy as np
import cython

ctypedef int index_t

cdef np.float64_t HALF_PLUS_EPSILON = 0.5 + 10.0 * np.finfo(np.float64).eps

cdef inline double my_fabs(double x) nogil:
    """
    simple inline fabs method
    """
    return x if x > 0 else -x

cdef inline int my_abs(int x) nogil:
    """
    simple inline abs method
    """
    return x if x > 0 else -x


cdef inline int my_round_index_t(double x) nogil:
    return <index_t>(x + HALF_PLUS_EPSILON)

cdef inline np.int16_t my_round_int16(double x) nogil:
    return <np.int16_t>(x + HALF_PLUS_EPSILON)

cdef inline size_t my_round_size_t(double x) nogil:
    return <size_t> (x + HALF_PLUS_EPSILON)

cdef inline void transform_point(np.float64_t[:, :] m,
                                 double *pi, double *po) nogil:
    po[0] = pi[0] * m[0][0] + pi[1] * m[0][1] + pi[2] * m[0][2] + m[0][3]
    po[1] = pi[0] * m[1][0] + pi[1] * m[1][1] + pi[2] * m[1][2] + m[1][3]
    po[2] = pi[0] * m[2][0] + pi[1] * m[2][1] + pi[2] * m[2][2] + m[2][3]

    
cdef inline void transform_vector(np.float64_t[:, :] m,
                                 double *vi, double *vo) nogil:
    vo[0] = vi[0] * m[0][0] + vi[1] * m[0][1] + vi[2] * m[0][2]
    vo[1] = vi[0] * m[1][0] + vi[1] * m[1][1] + vi[2] * m[1][2]
    vo[2] = vi[0] * m[2][0] + vi[1] * m[2][1] + vi[2] * m[2][2]

cdef inline double w_linear(double x) nogil:
    """
    Kernel for linear interpolation.
    
    Args:
        x: offset value, range [0, 1]
    
    Returns:
        kernel value
    """
    return 1.0 - my_fabs(x)
    
cdef inline bint is_within_image_border_1(double *p,
                                 np.int16_t[:, :, :] a) nogil:
    """Check if voxel coordinate is within image array
    
        Note that the access order of the array is a[z, y, x] whereas
        x is a vector (x, y, z)
    
        Args:
            x: numpy array containing the position
            a: grey value array
        Returns:
            True if x in inside the image excluding 1 voxel border,
            False otherwise
    """

    return p[0] >= 0.5 and p[1] >= 0.5 and p[2] >= 0.5 and \
        p[2] < a.shape[0] - 1.5 and p[1] < a.shape[1] - 1.5 and \
        p[0] < a.shape[2] - 1.5

cdef inline bint is_within_image(double *p,
                                 np.int16_t[:, :, :] a) nogil:
    """Check if voxel coordinate is within image array
    
        Note that the access order of the array is a[z, y, x] whereas
        x is a vector (x, y, z)
    
        Args:
            x: numpy array containing the position
            a: grey value array
        Returns:
            True if x in inside the image, False otherwise
    """

    return p[0] >= -0.5 and p[1] >= -0.5 and p[2] >= -0.5 and \
        p[2] < a.shape[0] - 0.5 and p[1] < a.shape[1] - 0.5 and \
        p[0] < a.shape[2] - 0.5

cdef inline void set_gm_environment(
        np.int16_t[:, :, :] in_im, np.int16_t[:, :, :] out_im,
        np.int32_t x, np.int32_t y, np.int32_t z, bint is_moving):
    cdef index_t x_d, y_d, z_d 
    cdef int delta_z, delta_y, delta_x

    if is_moving:
        for z_d in range(max(z-2, 0), min(z+3, out_im.shape[0])):
            for y_d in range(max(y-2, 0), min(y+3, out_im.shape[1])):
                for x_d in range(max(x-2, 0), min(x+3, out_im.shape[2])):
                    if out_im[z_d, y_d, x_d] == 0:
                        out_im[z_d, y_d, x_d] = -in_im[z_d, y_d, x_d]
    else:
        for z_d in range(max(z-2, 0), min(z+3, out_im.shape[0])):
            delta_z = my_abs(z_d-z)
            for y_d in range(max(y-2+delta_z, 0), 
                             min(y+3-delta_z, out_im.shape[1])):
                delta_y = my_abs(y_d-y)
                for x_d in range(max(x-2+delta_y+delta_z, 0), 
                                 min(x+3-delta_z-delta_y, out_im.shape[2])):
                    if out_im[z_d, y_d, x_d] == 0:
                        out_im[z_d, y_d, x_d] = -in_im[z_d, y_d, x_d]

cdef inline np.int16_t get_3d_array_5(np.int16_t *arr, int z, int y, 
        int x) nogil:
    return arr[x + y * 5 + z * 25]

cdef inline void set_3d_array_5(np.int16_t *arr, int z, int y, int x,
        np.int16_t val) nogil:
    arr[x + y * 5 + z * 25] = val

cdef inline np.float64_t get_3d_array_3(np.float64_t *arr, 
                                        int z, int y, int x) nogil:
    return arr[x + y * 3 + z * 9]

cdef inline void set_3d_array_3(np.float64_t *arr, int z, int y, int x,
        np.float64_t val) nogil:
    arr[x + y * 3 + z * 9] = val


cpdef void rescale_gvals(np.float64_t[:,:,:] in_im,
        np.int16_t[:,:,:] out_im, double g_min, double g_max,
        np.int16_t out_min, np.int16_t out_max, np.float64_t epsilon,
        bint is_mask)
cdef double mirrored_interpolate_1d(double *a, index_t N, double x) nogil
cpdef double mirrored_interpolate_1d_py(np.float64_t[:] a, double x) nogil

cdef double masked_interpolate_3d(np.int16_t[:, :, :] a, double *x) nogil
cpdef double masked_interpolate_3d_py(np.int16_t[:, :, :] a, \
        np.float64_t[:] x) nogil
cdef double mirrored_interpolate_3d(np.int16_t[:, :, :] a, double *x) nogil
cpdef double mirrored_interpolate_3d_py(np.int16_t[:, :, :] a, \
        np.float64_t[:] x) nogil

cpdef void downsample_XY(np.int16_t[:,:,:] in_im,\
                         np.float64_t[:, :, :] out_im,\
                         np.float64_t[:] scale_factor,\
                         np.float64_t[:] offset,\
                         int thread_num=*) nogil
cpdef void downsample_Z(np.float64_t[:,:,:] in_im,\
                        np.int16_t[:, :, :] out_im,\
                         np.float64_t[:] scale_factor,\
                        np.float64_t[:] offset,\
                        int thread_num=*,
                        bint is_mask=*) nogil
cpdef void downsample(np.int16_t[:,:,:] in_im,\
                      np.int16_t[:, :, :] out_im,\
                      int thread_num=*,
                      bint is_mask=*)
cpdef void downsample_with_offset(np.int16_t[:,:,:] in_im,\
                      np.int16_t[:, :, :] out_im,\
                      np.float64_t[:, :, :] tmp_im,\
                      np.float64_t[:] scale_factor,\
                      np.float64_t[:] offset,\
                      int thread_num=*,
                      bint is_mask=*)
cpdef void interpolate_block(np.int16_t[:,:,:] fixed_im,
                      np.int16_t[:, :, :] moving_im,
                      np.float64_t[:, :] t_matrix_fixed_to_moving,
                      int thread_num=*,
                      bint is_mask=*)
cpdef void fill_gm_image(np.int16_t[:,:,:] in_block, 
        np.int16_t[:,:,:] mask_block, np.float32_t[:,:,:] gm_block, 
        int thread_num)
cpdef np.float32_t gm_max(np.float32_t[:,:,:] gm_block) 
cpdef void fill_gm_histogram(np.float32_t[:,:,:] gm_block,
        np.uint64_t[:] hist, double gm_max)
cpdef void gm_masking(np.int16_t[:,:,:] in_block, 
        np.int16_t[:,:,:] out_im, np.float32_t[:,:,:] gm_block, 
        int gm_z_offset, double gm_limit, bint is_moving) 
cpdef void do_run_length_encoding(np.int16_t[:,:,:] im, int thread_num) nogil
cpdef void add_inertia_tensor_slice(np.int16_t[:, :] im,
        index_t z, np.float64_t[:, :] voxel_to_world_matrix, 
        np.int64_t[:] N, np.float64_t[:] centre, np.float64_t[:, :] TI)
cpdef void lcc_block( np.int16_t[:, :, :] fixed_im, 
        np.int16_t[:, :, :] moving_im, np.float64_t[:, :] t_mat_vox,
        np.int64_t[:] N_buf, np.float64_t[:] lcc_buf, np.int32_t thread_num,
        bint use_rle=*)
cpdef void lcc_block_with_derivatives( np.int16_t[:, :, :] fixed_im,
        np.int16_t[:, :, :] moving_im, np.float64_t[:, :] fixed_v2w_matrix,
        np.float64_t[:, :] trans_mat_vox, np.float64_t[:,:] trans_mat_gradient,
        np.int32_t n_p,
        np.int64_t[:] N_buf, np.float64_t[:] lcc_buf, 
        np.float64_t[:, :] gradient_buf, np.float64_t[:, :, :] hessian_buf,
        np.int32_t thread_num, bint use_rle=*)
cpdef void remove_rle(np.int16_t[:, :] in_slice, np.int16_t[:, :] out_slice)



if __name__ == "__main__":
    from setuptools import setup, find_packages, Extension
    from sys import platform

    try:
        from numpy.distutils.misc_util import get_numpy_include_dirs

        numpy_include_dirs = get_numpy_include_dirs()
    except ModuleNotFoundError:
        sys.exit("Please install numpy first")
    import lcreg

    if platform.startswith("win"):
        openmp_comp_args = [
            "/O2",
            "/openmp",
        ]
        openmp_link_args = [
            "/openmp",
        ]
    elif platform.startswith("linux"):
        openmp_comp_args = [
            "-DNDEBUG",
            "-O3",
            "-fopenmp",
        ]
        openmp_link_args = ["-fopenmp"]
    elif platform.startswith("darwin"):
        openmp_comp_args = [
            "-DNDEBUG",
            "-O3",
            "-fopenmp",
        ]
        openmp_link_args = ["-fopenmp", "-lgomp"]
    else:
        print("unsupported platform, please edit setup.py")
        exit(0)

    with open("README.md", "r") as fh:
        long_description = fh.read()

    setup(
        name="lcreg",
        version=lcreg.__version__,
        description="Efficient 3D rigid and affine image registration",
        long_description_content_type="text/markdown",
        long_description=long_description,
        url="https://github.com/p-roesch/lcreg",
        author="Peter Rösch",
        author_email="lcreg@hs-augsburg.de",
        license="GPLv3",
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            "Environment :: Console",
            "Intended Audience :: Science/Research",
            "Programming Language :: Cython",
            "Programming Language :: Python :: 3 :: Only",
        ],
        keywords="3D image registration",
        project_urls={
            "ResearchGate Project": "https://www.researchgate.net/project/Efficient-registration-of-large-3D-images-lcreg",
        },
        packages=find_packages(exclude=["docs", "tests*"]),
        python_requires=">=3, <4",
        install_requires=[
            "numpy>=1.21",
            "scipy>=1.7",
            "bcolz>=1.2",
            "psutil>=5.8",
            "py-cpuinfo>=8.0",
        ],
        entry_points={
            "console_scripts": [
                "lcreg=lcreg.lcreg:main",
                "lcreg_profile=lcreg.lcreg:profiling_main",
                "view_compressed_images=lcreg.view_compressed_images:main",
                "compressed_to_mhd=lcreg.image3d:export_bcolz_image_main",
                "abs_difference_mhd=lcreg.image3d:abs_difference_mhd_main",
                "difference_mhd=lcreg.image3d:difference_mhd_main",
                "isq_to_mhd=lcreg.isq_to_mhd:main",
            ],
        },
        ext_modules=[
            Extension(
                "lcreg.lcreg_lib",
                ["lcreg/lcreg_lib.bycython.c"],
                extra_compile_args=openmp_comp_args,
                extra_link_args=openmp_link_args,
                include_dirs=numpy_include_dirs,
            ),
        ],
    )



# lcreg: Efficent rigid and affine 3D image registration
#
# Copyright (C) 2019  Peter Rösch, Peter.Roesch@hs-augsburg.de
#
# Organisation:
# Faculty of Computer Science, Augsburg University of Applied Sciences,
# An der Hochschule 1, 86161 Augsburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
    lcreg utility functions
"""
from __future__ import print_function, division
import ast
import psutil
import numpy as np


def array_to_string(arr):
    """
    Convert numpy array into 1D string.

    Args:
        arr (numpy array): numpy array
    Returns:
        str containing space separated numbers corresponding to
        the flattened array
    """
    if not isinstance(arr, np.ndarray):
        arr = np.array(arr)
    string = str(arr.flatten())[1:-1].strip()
    return " ".join(string.split())


def string_to_array(string, dtype=np.float64):
    """
    Convert string in to numpy array.

    Args:
        string (str): String containing space-separated numbers
        dtype (numpy dtype): dtype of numpy array to be created
    Returns:
        numpy array from string with requested dtype

    """
    string = " ".join(string.split())
    arr_str = "[" + string.replace(" ", ",") + "]"
    return np.array(ast.literal_eval(arr_str), dtype=dtype)


def system_nr_of_cores():
    """
    Get number of CPU cores
    """
    return psutil.cpu_count()


def system_memory_available_mb():
    """
    Get 50% of available memory in MB
    """
    mem_limit_mb = int(psutil.virtual_memory().available // 2 ** 21)
    return mem_limit_mb


def corners_vox_x_y_z(im):
    """
    Generate corner positions of an image in voxel coordinates

    Args:
        im (Image3D or numpy array): image
    """
    for z in (0, im.shape[0] - 1):
        for y in (0, im.shape[1] - 1):
            for x in (0, im.shape[2] - 1):
                yield (x, y, z)


def image_corners_vox(im):
    """
    Create matrix containing image edges in voxel coordinates

    Args:
        im (Image3d): image
    """
    corner_list = []
    for x, y, z in corners_vox_x_y_z(im):
        corner_list.append([x, y, z, 1])
    corners_vox = np.array(corner_list, dtype=np.float64).T
    return corners_vox


def image_corners_world(im):
    """
    Create matrix containing image edges in voxel coordinates

    Args:
        im (Image3d): image
    """
    corners_world = np.dot(im.voxel_to_world_matrix, image_corners_vox(im))
    return corners_world



# lcreg: Efficent rigid and affine 3D image registration
#
# Copyright (C) 2019  Peter Rösch, Peter.Roesch@hs-augsburg.de
#
# Organisation:
# Faculty of Computer Science, Augsburg University of Applied Sciences,
# An der Hochschule 1, 86161 Augsburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
    Levenberg-Marquardt and simple gradient optimisation
"""

import logging
import ast
from itertools import permutations
import numpy as np
from lcreg import image3d
from lcreg import similarity
from lcreg.image_transform import t_matrix_world_from_p, p_from_t_matrix_world
from lcreg.lcreg_util import image_corners_world, corners_vox_x_y_z
from lcreg.image_transform import euler_from_matrix
import lcreg.lcreg_lib as cy_ip


def inertia_tensor(im):
    N = np.zeros(1, dtype=np.int64)
    centre = np.zeros(3, dtype=np.float64)
    TI = np.zeros((3, 3), dtype=np.float64)
    for z in range(1, im.shape[0] - 1):
        cy_ip.add_inertia_tensor_slice(
            im.data[z, :, :], z, im.voxel_to_world_matrix, N, centre, TI
        )
    # fill lower part of symmetric tensor
    for i in range(0, 3):
        for k in range(i + 1, 3):
            TI[k, i] = TI[i, k]
    centre /= N[0]
    I3 = np.identity(3, dtype=np.float64)
    TI -= N[0] * (np.dot(centre, centre) * I3 - np.outer(centre, centre))
    return centre, TI


def get_initial_transform_from_moments(reg_config):
    """
    Calculate intial transform from inertia tensors of masked images.
    """
    # find and open masked images in the middle of the pyramid
    fixed_masked_list_str = reg_config["optimisation"]["fixed_masked_images"]
    fixed_masked_list = ast.literal_eval(fixed_masked_list_str)
    im_index = len(fixed_masked_list) // 2
    fixed_im_name = fixed_masked_list[im_index]
    fixed_im = image3d.open_image(fixed_im_name)
    moving_masked_list_str = reg_config["optimisation"]["moving_masked_images"]
    moving_masked_list = ast.literal_eval(moving_masked_list_str)
    im_index = len(moving_masked_list) // 2
    moving_im_name = moving_masked_list[im_index]
    moving_im = image3d.open_image(moving_im_name)

    # calculate inertia tensor slice by slice
    c_fixed, TI_fixed = inertia_tensor(fixed_im)
    c_moving, TI_moving = inertia_tensor(moving_im)
    # calculate eigenvalues and eigenvectors
    evect_fixed = np.linalg.eigh(TI_fixed)[1]
    evect_moving = np.linalg.eigh(TI_moving)[1]
    # start with identity transform
    p_opt = np.zeros(6, dtype=np.float64)
    lcc_opt = similarity.lcc(fixed_im, moving_im, p_opt, False)[1]
    logging.debug("lcc identity transform: " + str(lcc_opt))
    p_shift = p_opt.copy()
    p_shift[:3] = c_moving[:] - c_fixed[:]
    lcc_shift = similarity.lcc(fixed_im, moving_im, p_shift, False)[1]
    logging.debug("lcc shift only:" + str(lcc_shift))
    if lcc_shift < lcc_opt:
        lcc_opt = lcc_shift
        p_opt[:] = p_shift[:]
    if not reg_config.getboolean("transformation", "initialise_shift_only"):
        # point set in fixed image
        pt_fixed = np.ones((4, 4), dtype=np.float64)
        for i in range(3):
            pt_fixed[:3, i] = c_fixed[:] + evect_fixed[:, i]
        pt_fixed[:3, 3] = c_fixed[:]
        pt_fixed_inv = np.linalg.inv(pt_fixed)
        pt_moving = np.ones((4, 4), dtype=np.float64)
        pt_moving[:3, 3] = c_moving[:]
        #
        # try all permutations and signs for moving image points
        result_list = []
        s = [1, 1, 1]
        for a in permutations((0, 1, 2), 3):
            for s[0] in (-1, 1):
                for s[1] in (-1, 1):
                    for s[2] in (-1, 1):
                        for i in range(3):
                            pt_moving[:3, i] = (
                                c_moving[:] + s[i] * evect_moving[:, a[i]]
                            )
                            M = np.dot(pt_moving, pt_fixed_inv)
                            # check for real rotation
                            if abs(np.linalg.det(M[:3, :3]) - 1) < 1e-3:
                                p = p_from_t_matrix_world(M)
                                N, lcc = similarity.lcc(
                                    fixed_im, moving_im, p, False
                                )
                                result_list.append((N, lcc, p))
        # ignore solutions with very small overlap
        N_max = max([t[0] for t in result_list if t[0] is not None])
        N_limit = int(N_max * 0.75)
        # find settings corresponding to minimum lcc
        for N, lcc, p in result_list:
            if N is not None and N >= N_limit and lcc < lcc_opt:
                lcc_opt, p_opt = lcc, p

    # store result in reg_config
    m_opt = t_matrix_world_from_p(p_opt)
    reg_config["optimisation"]["initial_transform"] = (
        repr(m_opt)
        .replace("array", "np.array")
        .replace("float64", "np.float64")
    )
    reg_config["optimisation"]["rotation_centre"] = (
        repr(c_fixed)
        .replace("array", "np.array")
        .replace("float64", "np.float64")
    )
    reg_config["optimisation"]["rotation_axes"] = (
        repr(evect_fixed)
        .replace("array", "np.array")
        .replace("float64", "np.float64")
    )
    logging.debug("initial_transform from moments: \n" + str(m_opt))
    logging.debug(
        "initial euler angles (xyz, deg):\n" + str(euler_from_matrix(m_opt))
    )
    logging.debug("initial translation vector:\n" + str(m_opt[:3, 3]))
    logging.debug("initial lcc value: " + str(lcc_opt))
    image3d.close_image(fixed_im)
    image3d.close_image(moving_im)


def __image_centre_world(im):
    """
    Image centre in world coordinates

    Args:
        im (Image3D): image
    """
    return im.transform_voxel_to_world((im.shape_array_xyz[::-1] - 1) / 2)


def __delta_r_from_delta_v(im, delta_v, c_rot):
    """
    Calculate approxiate gradient descent step with for rotation parameters.

    It is assumed that delta_v corresponds to small rotation angles.

    Args:
        im (Image3D): image
        delta_v (float): step with in units of image voxel diagonals
            for image corners
        c_rot (numpy array): centre of rotation

    Returns:
        rotation parameter step width for gradient descent

    """
    # centre of image in world coordinates
    r_w_max = -1
    for x, y, z in corners_vox_x_y_z(im):
        # find image corner with largest distance to the centre
        # of the coordinate system (world coordinates)
        p_w = im.transform_voxel_to_world(
            np.array([x, y, z], dtype=np.float64)
        )
        r_w = np.linalg.norm(p_w - c_rot)
        if r_w > r_w_max:
            r_w_max = r_w
    # assume that r_x = r_y = r_z for rotation parameters
    delta_r = delta_v * np.min(im.spacing) / r_w_max
    return delta_r


def __delta_v_from_delta_M(moving_im, M, delta_M, R_c_w):
    """
    Calculate maximum additional displacement of image edges resulting from
    transformation update.

    Args:
        moving_im (Image3D): moving image
        M (numpy array, 4x4): Transformation matrix
        delta_M (numpy array, 4x4): Transformation update matrix
        R_c_w: (numpy.array, 4x8): Matrix containing image edges
            in world coordinates
    """

    M_delta = np.dot(delta_M - np.identity(4), M)
    d_world = np.dot(M_delta, R_c_w)
    d_vox = np.dot(moving_im.world_to_voxel_matrix, d_world)
    d_vox_abs = np.linalg.norm(d_vox[:3, :], axis=0)
    return max(d_vox_abs)


def __affine_matrix_from_p(p):
    """
    Create affine transformation matrix from a vector with 12 entries.

    Args:
        p (numpy.array, 1x12, np.float64): parameter vector

    Returns:
        numpy.array, 4x4, numpy.float64; transformation matrix
    """
    M = np.identity(4, dtype=np.float64)
    M[:3, :] = p.reshape(3, 4)
    return M


def simple_gradient(
    fixed_im,
    moving_im,
    p_initial,
    c_rot=None,
    v_rot=None,
    d_v_initial=2,
    d_v_min=1,
    mem_limit_mb=-1,
    num_threads=-1,
):
    """
    Perform simple gradient descent as described in
        Studholme, C., Hill, D.L.G., Hawkes, D.J.:
        Automated 3-D registration of MR and CT images of the head.
        Med. Image Anal. 1 (1996) 163--175

    Args:
        fixed_im (Image3D): fixed image
        moving_im(Image3D): moving image
        p_initial (numpy.array): initial rigid (6) parameters
        c_rot (numpy.array): rotation centre, image centre if None is passed
        v_rot (numpy.array, 3x3): rotation axes, identity if None is passed
        d_v_initial: initial step width in units of moving_im voxel diameters.
        d_v_min: minimum step width in units of moving_im voxel diameters.
        mem_limit_mb (int): memory limit for optimisation in MB
        num_threads (int): number of threads to be used or -1 (use all cores)

    Returns:
        tuple (lcc_min, p_opt, n_steps):
            lcc_min: minimum (1-LCC) value
            p_opt: parameter values yielding minimum lcc
            n_steps: number of optimisation steps performed
    """
    if len(p_initial) != 6:
        err_str = "simple gradient optimisation requires rigid transform"
        logging.error(err_str)
        raise NotImplementedError(err_str)

    if c_rot is None:
        c_rot = __image_centre_world(fixed_im)
    if v_rot is None:
        v_rot = np.identity(3, dtype=np.float64)
    # matrix to shift rotation centre to origin
    T_c = np.identity(4, dtype=np.float64)
    T_c[:3, 3] = -c_rot
    # ... and back
    T_c_inv = np.identity(4, dtype=np.float64)
    T_c_inv[:3, 3] = c_rot
    # parameter step size
    delta_t = d_v_initial * np.linalg.norm(moving_im.spacing)
    delta_t_min = d_v_min * np.linalg.norm(moving_im.spacing)
    delta_r = __delta_r_from_delta_v(moving_im, d_v_initial, c_rot)
    # initial lcc value
    N, lcc_min = similarity.lcc(
        fixed_im, moving_im, p_initial, False, mem_limit_mb, num_threads
    )
    if N is None or lcc_min is None:
        logging.error("Insufficient overlap")
        return (None, None)

    # parameter offset
    offset = np.zeros(6, dtype=np.float64)
    # initial transformation matrix
    T_initial = t_matrix_world_from_p(p_initial)
    # optimum offset matrix
    d_T_opt = np.identity(4, dtype=np.float64)
    # optimum offset matrix for current step
    d_T_opt_step = np.identity(4, dtype=np.float64)
    # transformation parameters
    p_l = p_initial.copy()

    n_steps = 0
    logging.debug("lcc_initial: " + str(lcc_min))
    while delta_t >= delta_t_min:
        lcc_min_step = lcc_min
        # loop over parameters
        sign = 1
        for k in range(12):
            offset.fill(0)
            if k < 6:
                # translation
                offset[:3] = sign * delta_t * v_rot[:, (k // 2) % 3]
            else:
                # rotation
                offset[3:] = sign * delta_r * v_rot[:, (k // 2) % 3]

            # transformation offset
            d_T = np.dot(t_matrix_world_from_p(offset), d_T_opt)
            # current transformation matrix
            T_l = np.dot(
                T_initial, np.dot(T_c_inv, np.dot(np.linalg.inv(d_T), T_c))
            )
            p_l = p_from_t_matrix_world(T_l)
            N, lcc = similarity.lcc(
                fixed_im, moving_im, p_l, False, mem_limit_mb, num_threads
            )
            if N is not None and lcc is not None:
                if lcc < lcc_min_step:
                    lcc_min_step = lcc
                    d_T_opt_step[:, :] = d_T[:, :]
            sign *= -1

        if lcc_min_step < lcc_min:
            lcc_min = lcc_min_step
            # update optimum offset value
            d_T_opt[:, :] = d_T_opt_step[:, :]
            logging.debug(
                "lcc_min: {:f}, d_p_opt: {:s}".format(
                    lcc_min, str(list(p_from_t_matrix_world(d_T_opt)))
                )
            )
        else:
            delta_t = delta_t / 2.0
            delta_r = delta_r / 2.0
        n_steps += 1

    # get parameters corresponding to d_T_opt
    T_opt = np.dot(
        T_initial, np.dot(T_c_inv, np.dot(np.linalg.inv(d_T_opt), T_c))
    )
    p_opt = p_from_t_matrix_world(T_opt)

    return lcc_min, p_opt, n_steps


def _update_p(moving_im, R_c_w, p, grad, H, l, p_new):
    """
    Update parameter vector using LM algorithm

    Args:
        moving_im (Image3D) moving image
        R_c_w (np.array, 4x8, float64): Corners of fixed
                                image in world coordinates
        p (np.array, 1x6 or 1x12, float64): current parameter vector
        grad (np.array, 1x6 or 1x12, float64): gradient vector
        H (np.array, 6x6 or 12x12): Hessian matrix
        l (float64): LM parameter lambda
        p_new (np.array, 1x6 or 1x12, float64): Vector to store result in
    Returns:
        Shift of image corners in units of voxel diagonals 
        resulting from parameter update
    """

    # LM step
    D = np.diag(np.diag(H))
    delta_p = -np.dot(np.linalg.inv(H + l * D), grad)
    # update parameter vector
    if len(p) == 6:
        # rigid transformation
        M_offset_inv = np.linalg.inv(t_matrix_world_from_p(delta_p))
        M = t_matrix_world_from_p(p)
        M_new = np.dot(M, M_offset_inv)
        p_from_t_matrix_world(M_new, p_new, False)
    else:
        # affine transformation
        M_offset_inv = t_matrix_world_from_p(-delta_p)
        M = t_matrix_world_from_p(p)
        p_new[:] = p - delta_p
    return __delta_v_from_delta_M(moving_im, M, M_offset_inv, R_c_w)


def levenberg_marquardt(
    fixed_im,
    moving_im,
    p_initial,
    l_initial=1.0,
    l_downscale_factor=1.0 / 10.0,
    l_upscale_factor=10,
    max_step_num=50,
    min_vox_delta=0.1,
    max_nr_of_upscale_steps=4,
    mem_limit_mb=-1,
    num_threads=-1,
):
    """
    Perform Levenberg-Marquardt optimisation of transformation parameters

    Args:
        fixed_im (Image3D): fixed image
        moving_im(Image3D): moving image
        p_initial (numpy.array): initial rigid (6) or affine (12) parameters
        l_inital (float): Initial value for LM lambda parameter
        l_downscale_factor (float): lambda is multiplied with this factor
            after a step which led to a decrease of lcc
        l_upscale_factor (float): lambda is multiplied with this factor
            after a step which led to an increase of lcc
        max_step_num (int): maximum number of optimisation steps before
            optimisation terminates
        min_vox_delta (float): optimisation terminates if the parameter
            update resulting in a decrease of lcc corresponds to a
            maximum displacement of image corners smaller than min_vox_delta
        max_nr_of_upscale_steps (int): maximum number of steps resulting in
            lcc increase before optimisation terminates
        mem_limit_mb (int): memory limit for optimisation in MB
        num_threads (int): number of threads to be used or -1 (use all cores)

    Returns:
        tuple (lcc_min, p_opt, n_steps):
            lcc_min: minimum (1-LCC) value
            p_opt: parameter values yielding minimum lcc
            n_steps: number of optimisation steps performed
    """
    R_c_w = image_corners_world(fixed_im)
    l = l_initial
    p = p_initial.copy()
    p_opt = p_initial.copy()

    # initialise values
    lcc_val_opt, grad_opt, H_opt = similarity.lcc(
        fixed_im, moving_im, p_opt, True, mem_limit_mb, num_threads
    )[1:]
    logging.debug("initial lcc: {:f}, lambda: {:e}".format(lcc_val_opt, l))
    # first LM step
    _update_p(moving_im, R_c_w, p_opt, grad_opt, H_opt, l, p)
    nr_of_upscale_steps = 0

    is_finished = False
    # optimisation loop
    step_opt = 0
    for step in range(1, max_step_num + 1):
        lcc_val, grad, H = similarity.lcc(
            fixed_im, moving_im, p, True, mem_limit_mb, num_threads
        )[1:]
        logging.debug(
            "step: {:d} lcc: {:f}, lambda: {:e}".format(step, lcc_val, l)
        )
        #
        downhill = lcc_val <= lcc_val_opt
        if downhill:
            lcc_val_opt = lcc_val
            step_opt = step
            l = l * l_downscale_factor
            nr_of_upscale_steps = 0
            p_opt = p[:]
            grad_opt[:] = grad[:]
            H_opt[:, :] = H[:, :]
        else:
            l = l * l_upscale_factor
            nr_of_upscale_steps += 1
        #
        # parameter update
        delta_v = _update_p(moving_im, R_c_w, p_opt, grad_opt, H_opt, l, p)
        #
        # check stop conditions
        if step > 1:
            if min_vox_delta is not None:
                if downhill and delta_v < min_vox_delta:
                    logging.debug(
                        "LM termination: delta_v: {:f}".format(delta_v)
                    )
                    is_finished = True
            if nr_of_upscale_steps >= max_nr_of_upscale_steps:
                logging.debug(
                    "LM termination: {:d} upscale steps".format(
                        nr_of_upscale_steps
                    )
                )
                is_finished = True
        if is_finished:
            break
    return (lcc_val_opt, p_opt, step_opt)



# lcreg: Efficent rigid and affine 3D image registration
#
# Copyright (C) 2019  Peter Rösch, Peter.Roesch@hs-augsburg.de
#
# Organisation:
# Faculty of Computer Science, Augsburg University of Applied Sciences,
# An der Hochschule 1, 86161 Augsburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
    Similarity measure calculation (Local Correlation Coefficient)
"""
from __future__ import print_function, division
import math
import numpy as np

from lcreg import image3d
from lcreg.lcreg_lib import lcc_block, lcc_block_with_derivatives
from lcreg.lcreg_util import system_memory_available_mb
from lcreg.image_transform import required_transform_mem_mb
from lcreg.image_transform import transformed_offset_shape
from lcreg.image_transform import t_matrix_world_from_p, t_matrix_vox_from_p


def __gradient_matrix(img):
    """
        Create Matrix to transform gradients from voxel to world coordinates

        Args:
            img (Image3D): 3D image

        Returns:
            numpy.array (3x3, numpy.float64): gradient transformation matrix
    """
    return np.dot(img.transform_matrix, np.diag(0.5 / img.spacing))


def __create_N_buf(fixed_im):
    """
        Create empty accumulation buffer for voxel pair counter (N)

        Args:
            fixed_im (Image3D): fixed image

        Returns:
            numpy.array (1xfixed_im.shape[0], numpy.int64): accumulation buffer
    """
    return np.zeros(shape=(fixed_im.shape[0],), dtype=np.int64)


def __create_lcc_buf(fixed_im):
    """
        Create empty accumulation buffer for local correlation sums (lcc)

        Args:
            fixed_im (Image3D): fixed image

        Returns:
            numpy.array (1xfixed_im.shape[0], numpy.float64): Buffer
    """
    return np.zeros(shape=(fixed_im.shape[0],), dtype=np.float64)


def __create_gradient_buf(fixed_im, p):
    """
        Create empty accumulation buffer for lcc gradient sums

        Args:
            fixed_im (Image3D): fixed image
            p (numpy.array, numpy.float64): 6 (rigid) or 12 (affine) parameters

        Returns:
            numpy.array (fixed_im.shape[1]xlen(p), numpy.float64): Buffer
    """
    return np.zeros(shape=(fixed_im.shape[0], len(p)), dtype=np.float64)


def __create_hessian_buf(fixed_im, p):
    """
        Create empty accumulation buffer for lcc approximated hessian sums

        Args:
            fixed_im (Image3D): fixed image
            p (numpy.array, numpy.float64): 6 (rigid) or 12 (affine) parameters

        Returns:
            numpy.array (fixed_im.shape[0] x len(p) x len(p), numpy.float64):
                Buffer
    """
    return np.zeros(
        shape=(fixed_im.shape[0], len(p), len(p)), dtype=np.float64
    )


def __calc_lcc(
    fixed_im,
    moving_im,
    p,
    with_derivatives,
    N_sum,
    lcc_sum,
    gradient_sum,
    hessian_sum,
    thread_num=-1,
    use_rle=True,
):
    """
        Calculate the LCC measure with or without derivatives
            (Python interface)

        This function calls the optimised lcc_block functions implemented
        in cython.

        Args:
            fixed_im (Image3D): fixed image
            moving_im (Image3D): moving image
            p (numpy array, numpy.float64): rigid or affine parameter vector
            with_derivatives (bool): Flag, True if gradient and hessian
                have to be calculated.
            N_sum: (numpy array, numpy.int64, 1 entry): accumulation buffer
                for number of voxels used for LCC calculation
            lcc_sum: (numpy array, numpy.float64, 1 entry): accumulation buffer
                for lcc=(1-LCC) value
            gradient_sum: (numpy array, numpy.float64, 1xlen(p)):
                accumulation buffer for lcc gradient
            hessian_sum: (numpy array, numpy.float64, len(p)xlen(p)):
                accumulation buffer for lcc hessian
            thread_num: Number of threads to be used for calculation or
                -1 to use all cores
            use_rle (bool): use or ignore run length encoding

        Returns:
            tuple (N, lcc, gradient, hessian):
                N: Number of voxels used for LCC calculation
                lcc: Value of (1-LCC)
                if with_derivatives is True:
                    gradient: 1D numpy array with LCC derivatives w.r.t. p
                    hessian: 2D numpy array with aprrox. LCC hessian w.r.t. p
                if not enough corresponding masked voxels could be found,
                    (None, None) is returned
    """
    fixed_data = fixed_im.data[:, :, :]
    moving_data = moving_im.data[:, :, :]
    trans_mat_vox = t_matrix_vox_from_p(fixed_im, moving_im, p)
    gradient_matrix = __gradient_matrix(fixed_im)
    N_buf = __create_N_buf(fixed_im)
    lcc_buf = __create_lcc_buf(fixed_im)
    if with_derivatives:
        gradient_buf = __create_gradient_buf(fixed_im, p)
        hessian_buf = __create_hessian_buf(fixed_im, p)
        lcc_block_with_derivatives(
            fixed_data,
            moving_data,
            fixed_im.voxel_to_world_matrix,
            trans_mat_vox,
            gradient_matrix,
            len(p),
            N_buf,
            lcc_buf,
            gradient_buf,
            hessian_buf,
            thread_num,
            use_rle,
        )
    else:
        lcc_block(
            fixed_data,
            moving_data,
            trans_mat_vox,
            N_buf,
            lcc_buf,
            thread_num,
            use_rle,
        )
    N_sum[0] += N_buf.sum()
    lcc_sum[0] += lcc_buf.sum(0)
    del N_buf
    del lcc_buf
    if with_derivatives:
        gradient_sum += gradient_buf.sum(0)
        hessian_sum += hessian_buf.sum(0)
        del gradient_buf
        del hessian_buf


def lcc(
    fixed_im,
    moving_im,
    p,
    with_derivatives=True,
    mem_limit_mb=-1,
    num_threads=-1,
    use_rle=True,
):
    """
    Perform block wise lcc calculation of an image pair.

    Args:
        fixed_im (Image3D): fixed image
        moving_im (Image3D): moving image
        p (numpy float64 array): 6 (rigid) or 12 (affine)
            transformation parameters
        with_derivatives (bool): True to calculate gradient and hessian
        mem_limit_mb (int): memory limit for local buffers in MB
        num_threads (int): number of threads to be used (-1: use all cores)
        use_rle (bool): use run lenght encoding

    Returns:
        tuple with (N, lcc, ) or (N, lcc, gradient, hessian) for
        with_derivatives==False or with_derivatives==True respectively
    """

    if mem_limit_mb < 0:
        mem_limit_mb = system_memory_available_mb() // 2

    if isinstance(fixed_im.data, np.ndarray):
        mem_limit_mb += fixed_im.mem_mb

    if isinstance(moving_im.data, np.ndarray):
        mem_limit_mb += moving_im.mem_mb

    lcc_sum = np.zeros(1, dtype=np.float64)
    N_sum = np.zeros(1, dtype=np.int64)
    if with_derivatives:
        gradient_sum = np.zeros(len(p), dtype=np.float64)
        hessian_sum = np.zeros((len(p), len(p)), dtype=np.float64)
    else:
        gradient_sum = None
        hessian_sum = None
    if fixed_im.mem_mb + moving_im.mem_mb <= mem_limit_mb:
        __calc_lcc(
            fixed_im,
            moving_im,
            p,
            with_derivatives,
            N_sum,
            lcc_sum,
            gradient_sum,
            hessian_sum,
            num_threads,
            use_rle,
        )
    else:
        use_rle = False
        block_size = np.array((100, 100, 100), dtype=np.int32)
        mem_100 = required_transform_mem_mb(
            block_size, fixed_im, moving_im, t_matrix_world_from_p(p)
        )
        factor = math.pow(mem_limit_mb / mem_100, 1 / 3)
        block_size = (factor * block_size).astype(np.int32)
        # subtract overlap region for both sides of the image
        overlap_size = 2
        for z in range(0, fixed_im.shape[0], block_size[2] - 2 * overlap_size):
            s_z = min(fixed_im.shape[0] - z, block_size[2])
            for y in range(
                0, fixed_im.shape[1], block_size[1] - 2 * overlap_size
            ):
                s_y = min(fixed_im.shape[1] - y, block_size[1])
                for x in range(
                    0, fixed_im.shape[2], block_size[0] - 2 * overlap_size
                ):
                    s_x = min(fixed_im.shape[2] - x, block_size[0])
                    if min(s_x, s_y, s_z) > 0:
                        fixed_edges = np.array(
                            [
                                [x, y, z],
                                [x + s_x, y, z],
                                [x, y + s_y, z],
                                [x, y, z + s_z],
                                [x + s_x, y + s_y, z],
                                [x + s_x, y, z + s_z],
                                [x, y + s_y, z + s_z],
                                [x + s_x, y + s_y, z + s_z],
                            ],
                            dtype=np.float64,
                        ).T
                        moving_offset, moving_shape = transformed_offset_shape(
                            fixed_im,
                            moving_im,
                            fixed_edges,
                            t_matrix_world_from_p(p),
                        )
                        if min(moving_shape) > 0:
                            fixed_part = image3d.image_part(
                                fixed_im,
                                np.array((x, y, z)),
                                np.array((s_x, s_y, s_z)),
                            )
                            moving_part = image3d.image_part(
                                moving_im, moving_offset, moving_shape
                            )
                            __calc_lcc(
                                fixed_part,
                                moving_part,
                                p,
                                with_derivatives,
                                N_sum,
                                lcc_sum,
                                gradient_sum,
                                hessian_sum,
                                num_threads,
                                use_rle,
                            )
                            del fixed_part
                            del moving_part
    N = N_sum[0]
    if N == 0:
        result = (None, None)
    else:
        if with_derivatives:
            # fill hessian (symmetry ...)
            for u in range(len(p)):
                for v in range(u + 1, len(p)):
                    hessian_sum[v, u] = hessian_sum[u, v]
            result = (N, lcc_sum[0] / 2 / N, gradient_sum / N, hessian_sum / N)
        else:
            result = (N, lcc_sum[0] / N)
    return result
